#include <cstdint>
#include <functional>
#include <iostream>
#include <vector>

namespace Fiber
{
    // 64KB stack
    const unsigned STACK_SIZE = 65536;

    // Thread context
    struct Context
    {
        intptr_t m_StackPointer;
    };
    static Context g_MainContext; // Stores the original context
    static Context* g_CurrentContext = &g_MainContext;  // Defaults to the original context

    struct Fiber
    {
        Context m_Context;
        std::function<void()> m_EntryPoint;
        uint8_t m_Stack[STACK_SIZE];

        struct State
        {
            enum Enum
            {
                None,
                Started,
                Ended
            };
        };
        State::Enum m_State;
    };

    // Define as global to use data section of memory
    static std::vector< std::shared_ptr<Fiber> > g_Fibers;

    void Schedule();

    // switch to a specific context
    void SwitchContext(Fiber* newFiber)
    {
        // save current context
        uintptr_t stackPointer;
        _asm
        {
            pushad;
            mov stackPointer, esp;
        }
        g_CurrentContext->m_StackPointer = stackPointer;

        g_CurrentContext = newFiber ? &newFiber->m_Context : &g_MainContext;
        stackPointer = g_CurrentContext->m_StackPointer;

        // switch stack
        _asm
        {
            mov esp, stackPointer;
        }

        if (newFiber && newFiber->m_State == Fiber::State::None)
        {
            // keep the fiber pointer on the stack so we can get it later
            _asm
            {
                mov eax, newFiber;
                push eax;
            }
            newFiber->m_State = Fiber::State::Started;
            newFiber->m_EntryPoint();
            _asm
            {
                pop eax;
                mov newFiber, eax;
            }
            newFiber->m_State = Fiber::State::Ended;
            Schedule(); // leave this fiber behind
        }

        // restore context
        _asm
        {
            popad;
        }
    }

    // Create a new context 
    std::shared_ptr<Fiber> CreateFiber(std::function<void()> entryPoint)
    {
        auto fiber = std::make_shared<Fiber>();
        fiber->m_Context.m_StackPointer = (intptr_t)&fiber->m_Stack[STACK_SIZE - 1];
        fiber->m_EntryPoint = entryPoint;
        fiber->m_State = Fiber::State::None;
        std::memset(fiber->m_Stack, 0, sizeof(fiber->m_Stack));
        return fiber;
    }

    // Switch from one context to another
    void Schedule()
    {
        // Simple scheduler that just round robins the fibers
        static unsigned nextFiberToSchedule = 0;
        unsigned fiberToSchedule = g_Fibers.size();
        for (unsigned i = 0; i < g_Fibers.size(); ++i)
        {
            unsigned currentFiberIndex = (nextFiberToSchedule + i) % g_Fibers.size();
            if (g_Fibers[currentFiberIndex]->m_State == Fiber::State::Ended)
                continue;
            fiberToSchedule = currentFiberIndex;
            break;
        }

        // By default switch to the main context
        Fiber* newFiber = nullptr;
        // If we have anything left to schedule, schedule that first
        if (fiberToSchedule != g_Fibers.size())
        {
            // Update next fiber to schedule
            nextFiberToSchedule = (fiberToSchedule + 1) % g_Fibers.size();
            newFiber = g_Fibers[fiberToSchedule].get();
        }

        // Perform the actual context switch
        SwitchContext(newFiber);
    }

    // Returns the number of fibers scheduled since the thread was yielded
    unsigned Yield()
    {
        Schedule();
        return 0;
    }
}

// Driver code below

void Ascending(int count, int yieldEvery)
{
    for (int i = 0; i < count; ++i)
    {
        if (i > 0 && i % yieldEvery == 0)
        {
            std::cout << std::endl;
            Fiber::Yield();
        }
        std::cout << i << ", ";
    }
    std::cout << std::endl;
}

void Descending(int count, int yieldEvery)
{
    for (int i = 0; i < count; ++i)
    {
        if (i > 0 && i % yieldEvery == 0)
        {
            std::cout << std::endl;
            Fiber::Yield();
        }
        std::cout << (count - i) << ", ";
    }
    std::cout << std::endl;
}


void Start()
{
    // Restart all fiber execution
    Fiber::Schedule();
}

int main()
{
    // create one fiber for ascending and one fiber for descending
    Fiber::g_Fibers.push_back(Fiber::CreateFiber([]()
    {
        Ascending(100, 10);
    }));

    Fiber::g_Fibers.push_back(Fiber::CreateFiber([]()
    {
        Descending(100, 10);
    }));

    Start();

    return 0;
}
